@@ -1,5 +1,4 @@
-# Starfish - Perl-based System for Text-Embedded
-#     Programming and Preprocessing
+# Starfish - Perl-based System for Preprocessing and Text-Embedded Programming
 #
 # (c) 2001-2020 Vlado Keselj http://web.cs.dal.ca/~vlado vlado@dnlp.ca
 #               and contributing authors
@@ -20,7 +19,7 @@ use vars qw($VERSION @ISA @EXPORT @EXPORT_OK %EXPORT_TAGS); # Exporter vars
 %EXPORT_TAGS = ( 'all' => [qw(
   appendfile echo file_modification_date 
   file_modification_time getfile getmakefilelist get_verbatim_file
-  getinclude include
+  getinclude htmlquote include
   last_update putfile read_records read_starfish_conf
   sfish_add_tag sfish_ignore_outer starfish_cmd make_gen_dirs_to_generate
   make_add_dirs_to_generate_if_needed
@@ -31,11 +30,8 @@ use vars qw($VERSION @ISA @EXPORT @EXPORT_OK %EXPORT_TAGS); # Exporter vars
 # updated here and in META.yml
 use vars qw($NAME $ABSTRACT $VERSION);
 $NAME     = 'Starfish';
-$ABSTRACT = 'Perl-based System for Text-Embedded Programming and Preprocessing';
-$VERSION  = '1.31';
-
-use vars qw($Revision);
-($Revision = substr(q$Revision: $, 10)) =~ s/\s+$//;
+$ABSTRACT = 'Perl-based System for Preprocessing and Text-Embedded Programming';
+$VERSION  = '1.32';
 
 use vars qw(@DirGenerateIfNeeded);
 
@@ -70,6 +66,7 @@ sub new($@) {
 	$self->{Style}           = $::Star->{Style};
 	$self->{CodePreparation} = $::Star->{CodePreparation};
 	$self->{LineComment}     = $::Star->{LineComment};
+	$self->{OutDelimiters}   = $::Star->{OutDelimiters};
 	$self->{IgnoreOuter}     = $::Star->{IgnoreOuter};
 	$self->{hook}            = [ @{ $::Star->{hook} } ];
     }
@@ -231,6 +228,7 @@ sub eval_dispatch {
 	my $r = &{$self->{hook}->[$self->{ttype}]->{replace}}( $self, @{ $self->{args} } );
 	return $r if $self->{REPLACE};
 	return $c if $r eq '';
+	#!! This line needs to be adapted to use OutDelimiters
 	return $c.$self->{hook}->[0]->{'begin'}.$r.$self->{hook}->[0]->{'end'};
     }
     die;
@@ -366,26 +364,34 @@ sub scan {
 	}
 	
 	if ($self->{ttype}==-2) {$self->{currenttoken}=$self->{data}; $self->{data}=''}
-        elsif (@{$self->{args}}) {
-	  #$self->{Out} = $newOut;
-	  $self->_process_outer( substr($self->{data}, 0, $i1) );
-	  #$self->{Out} .= substr($self->{data}, 0, $i1);
-	  ##$self->{data} = $newData;
-	  $self->{prefix} = '';
-	  $self->{currenttoken} = substr($self->{data}, $i1+$pl, $i2-$i1-$pl);
-	  $self->{suffix} = '';
-	  $self->{data} = substr($self->{data}, $i2+$sl);
-	}
-	else {
-	  # just copy type -2
-	  # instead of returning as earlier, to
-	  # support negative look-back for prefix
-	  # $self->{Out} .= substr($self->{data}, 0, $i1);
-	  $self->_process_outer( substr($self->{data}, 0, $i1) );
-	  $self->{prefix} = substr($self->{data}, $i1, $pl);
-	  $self->{currenttoken} = substr($self->{data}, $i1+$pl, $i2-$i1-$pl);
-	  $self->{suffix} = substr($self->{data}, $i2, $sl);
-	  $self->{data} = substr($self->{data}, $i2+$sl);
+	else { # live code
+	  if (@{$self->{args}}) {
+	    #$self->{Out} = $newOut;
+	    $self->_process_outer( substr($self->{data}, 0, $i1) );
+	    #$self->{Out} .= substr($self->{data}, 0, $i1);
+	    ##$self->{data} = $newData;
+	    $self->{prefix} = '';
+	    $self->{currenttoken} = substr($self->{data}, $i1+$pl, $i2-$i1-$pl);
+	    $self->{suffix} = '';
+	    $self->{data} = substr($self->{data}, $i2+$sl);
+	  }
+	  else {
+	    # just copy type -2
+	    # instead of returning as earlier, to
+	    # support negative look-back for prefix
+	    # $self->{Out} .= substr($self->{data}, 0, $i1);
+	    $self->_process_outer( substr($self->{data}, 0, $i1) );
+	    $self->{prefix} = substr($self->{data}, $i1, $pl);
+	    $self->{currenttoken} = substr($self->{data}, $i1+$pl, $i2-$i1-$pl);
+	    $self->{suffix} = substr($self->{data}, $i2, $sl);
+	    $self->{data} = substr($self->{data}, $i2+$sl);
+	  }
+	  # Remove old output if it is there:
+	  if (defined($self->{OutDelimiters})) {
+	    my ($b1,$b2,$e1,$e2) = @{ $self->{OutDelimiters} };
+	    if ($self->{data} =~ /^\Q$b1\E(\d*)\Q$b2\E.*?\Q$e1\E\1\Q$e2\E/s) {
+	      $self->{data} = $'; }
+	  }
 	}
       }
 
@@ -427,9 +433,21 @@ sub evaluate {
     if ($self->{REPLACE}) { return $::O }
     #
     # $self->{hook}->[0] is reserved for output pieces
-    #
-    if ($::O ne '')
-    { $suf.="$self->{hook}->[0]->{'begin'}$::O$self->{hook}->[0]->{'end'}" }
+    # It should be changed to the use of OutDelimiters only
+    if ($::O ne '') {
+      my ($b, $e);
+      if (defined($self->{OutDelimiters})) {
+	my @d = @{ $self->{OutDelimiters} };
+	$b = $d[0].$d[1]; my $i; $e = $d[2].$d[3];
+	if (index($::O, $e) != -1) {
+	  while(1) { $i++; $e=$d[2].$i.$d[3]; last if index($::O, $e)==-1;
+            die "Problem finding ending delimiter!\n(O=$::O)" if $i > 1000000;}
+	  $b = $d[0].$i.$d[1];
+	}
+      } else { $b = $self->{hook}->[0]->{'begin'};
+	       $e = $self->{hook}->[0]->{'end'}; }
+	  $suf.=$b.$::O.$e;
+    }	  
     return "$pref$c$suf";
 }
 
@@ -450,7 +468,8 @@ sub evaluate_py {
     local $::Star = $self;
     local $::O = '';
     $self->eval1($code, 'snippet');
- 
+
+    #!! to update OutDelimiters
     if ($self->{REPLACE}) { return $::O }
     if ($::O ne '') {
 	my $indent = '';
@@ -483,13 +502,22 @@ sub evaluate_py1 {
     local $::O = '';
     local $::Star = $self;
     $self->eval1($code, 'snippet');
- 
+
     if ($self->{REPLACE}) { return $indent.$::O }
-    elsif ($::O eq '') { return "$indent#$prefix$c!>\n" }
+    elsif ($::O eq '') { return "$indent#$prefix$c!>" }
     else {
       $::O =~ s/^/$indent/gmx;
-      my $r= "$indent#$prefix$c!>\n$indent#+\n".$::O;
-      $r =~ s/\n?$/\n/; $r.="$indent#-\n"; } # no extra \n
+      my $r;
+      my ($b,$e); my @d = @{ $self->{OutDelimiters} };
+      $b = $d[0].$d[1]; my $i; $e = $d[2].$d[3];
+      if (index($::O, $e) != -1) {
+	while(1) { $i++; $e=$d[2].$i.$d[3]; last if index($::O, $e)==-1;
+	  die "Problem finding ending delimiter!\n(O=$::O)" if $i > 1000000;}
+	$b = $d[0].$i.$d[1];
+      }
+      $r= "$indent#$prefix$c!>$b".$::O;
+      $r =~ s/\n?$/\n/; $r.="$indent$e"; # no extra \n
+    }
 }
 
 # predefined evaluator: echo
@@ -500,6 +528,7 @@ sub eval_echo {
     my $suff = shift;
     $::O = $cont;
 
+    # to update OutDelimiters
     return $::O if $self->{REPLACE};
     return $pref.$cont.$suff if $::O eq '';
     $suff.="$self->{hook}->[0]->{'begin'}$::O$self->{hook}->[0]->{'end'}";
@@ -755,6 +784,21 @@ sub setGlobStyle {
     $self->setStyle($s);
 }
 
+sub clearStyle {
+  my $self = shift;
+  foreach my $k (qw(Style CodePreparation LineComment OutDelimiters
+                    IgnoreOuter)) {
+    delete $self->{$k} }
+  $self->{hook} = [];
+}
+
+# List of fields typically set in setStyle:
+# $self->{Style}           = string, style
+# $self->{CodePreparation} = scalar
+# $self->{LineComment}     = string, line comment
+# $self->{OutDelimiters}   = [] eg: "//" "+\n" "//" "-\n"
+# $self->{IgnoreOuter}     = 1 or ''
+# $self->{hook}            = [], list of hooks
 sub setStyle {
     my $self = shift;
     if ($#_ == -1) {
@@ -784,37 +828,57 @@ sub setStyle {
     if (defined $self->{Style} and $s eq $self->{Style}) { return }
     
     # default
+    $self->clearStyle();
     $self->{'LineComment'} = '#';
     $self->{'IgnoreOuter'} = '';
+    #!!should be introduced after all styles are fixed
+    #$self->{OutDelimiters} = [ "\n#", "+\n", "#", "-" ];
     $self->{hook}= [
-      {begin => "\n#+\n", end=>"\n#-\n", f=>sub{return''}}, # Reserved for output
-      {begin => '<?', end => '!>', f => \&evaluate },
+      {begin => '#<?', end => '!>', f => \&evaluate },
+      {begin => '<?',  end => '!>', f => \&evaluate },
       {begin => '<?starfish', end => '?>', f => \&evaluate }
     ];
     $self->{CodePreparation} = 's/\\n(?:#|%|\/\/+)/\\n/g';
 
     # Used for Python and Makefile with &evaluate_py1
-    my $re_py1 = qr/([\ \t]*)\#(\ *<\?)([\000-\377]*?)!>\n/x;
+    my $re_py1 = qr/([\ \t]*)\#(\ *<\?)([\000-\377]*?)!>/x;
     # extension below was a bug for <?...!>...<?...!>#+...#-
     # ([\ \t]*\#+\n[\000-\377]*?\n[\ \t]*\#-\n)?/x;
 
-    if ($s eq 'perl') { }
+    if ($s eq 'perl') {
+      #!! move later to default
+      $self->{OutDelimiters} = [ "#", "+\n", "#", "-" ];
+    }
     elsif ($s eq 'makefile') {
-	$self->{hook} = [
-	     {begin => qr/^\s*#\+\n/, end=>qr/\n\s*#-\n/, f=>sub{return""}}, # Reserved for output
-	    ];
-	$self->addHook($re_py1, \&evaluate_py1);
-	$self->{CodePreparation} = 's/\\n\\s*#/\\n/g';
+      $self->{OutDelimiters} = [ "#", "+\n", "#", "-" ];
+      $self->{hook} = [
+        # Reserved for output (!!to change to OutDelimiters)
+        #{begin => qr/^\s*#\+\n/, end=>qr/\n\s*#-\n/, f=>sub{return""}},
+      ];
+      $self->addHook($re_py1, \&evaluate_py1);
+      $self->{CodePreparation} = 's/\\n\\s*#/\\n/g';
+    }
+    elsif ($s eq 'python') {
+      $self->{OutDelimiters} = [ "#", "+\n", "#", "-" ];
+      $self->{hook} = [
+        # !!to change to OutDelimiters
+        #{begin => qr/^\s*#\+\n/, end=>qr/\n\s*#-\n/, f=>sub{return""}},
+      ];
+      $self->addHook($re_py1, \&evaluate_py1);
+      $self->{CodePreparation} = 's/\\n\\s*#/\\n/g';
     }
     elsif ($s eq 'java') {
-	$self->{LineComment} = '//';               #[0] reserved for output
-	$self->{hook} = [{begin => "//+\n", end=>"//-", f=>sub{return''}},
-          {begin => '//<?', end => '!>', f => \&evaluate },
-          {begin => '<?', end => '!>', f => \&evaluate }];
-	$self->{CodePreparation} = 's/^\s*\/\/+//mg';
+      $self->{LineComment} = '//';
+      $self->{OutDelimiters} = [ "//", "+\n", "//", "-" ];
+      #[0] reserved for output (!!to change to OutDelimiters)
+      $self->{hook} = [#!!toremove:{begin => "//+\n", end=>"//-", f=>sub{return''}},
+		       {begin => '//<?', end => '!>', f => \&evaluate },
+		       {begin => '<?', end => '!>', f => \&evaluate }];
+      $self->{CodePreparation} = 's/^\s*\/\/+//mg';
     }
     elsif ($s eq 'tex') {                          # TeX and LaTeX hooks
-	$self->{LineComment} = '%';
+      $self->{LineComment} = '%';
+      #!!needs update OutDelimiters
 	$self->{hook}=[{begin => "%+\n", end=>"\n%-\n", f=>sub{return''}},  # Reserved for output
 	       {begin => '%<?', end => "!>\n", f => \&evaluate },
 	       {begin => '<?', end => "!>\n", f => \&evaluate },
@@ -822,15 +886,11 @@ sub setStyle {
 
 	$self->{CodePreparation} = 's/^[ \t]*%//mg';
     }
-    elsif ($s eq 'python') {
-	$self->{hook} =
-	    [{begin => qr/^\s*#\+\n/, end=>qr/\n\s*#-\n/, f=>sub{return""}}, # Reserved for output
-	     ];
-	$self->addHook($re_py1, \&evaluate_py1);
-	$self->{CodePreparation} = 's/\\n\\s*#/\\n/g';
-    }
     elsif ($s eq '.html.sfish') {
-	undef $self->{LineComment};
+      undef $self->{LineComment};
+      #!!needs update OutDelimiters
+      #$self->{OutDelimiters} = [ "\n#", "+\n", "#", "-" ];
+
 	$self->{hook}=[{begin => "\n<!-- + -->\n", end=>"\n<!-- - -->\n",   # Reserved for output
 		f=>sub{return''}},
 	       {begin => '<!--<?', end => '!>-->', f => \&evaluate },
@@ -841,18 +901,18 @@ sub setStyle {
 	$self->{CodePreparation} = '';
     }
     elsif ($s eq 'html') {
-	undef $self->{LineComment};
-	$self->{hook}=[{begin => "\n<!-- + -->\n", end=>"\n<!-- - -->\n",   # Reserved for output
-		f=>sub{return''}},
-	       {begin => '<!--<?', end => '!>-->', f => \&evaluate },
-	       #{begin=>'<?', end=>'?>', f=>\&evaluate },
-	       {begin=>'<?starfish ', end=>'?>', f=>\&evaluate }
-        ];
-	$self->{CodePreparation} = '';
+      undef $self->{LineComment}; # Changes
+      $self->{OutDelimiters} = [ "<!-- +", " -->", "<!-- -", " -->" ];
+      $self->{hook}=[
+	{ begin => '<!--<?', end => '!>-->', f => \&evaluate },
+	{ begin=>'<?starfish ', end=>'?>', f=>\&evaluate }
+      ];
+      $self->{CodePreparation} = '';
     }
     elsif ($s eq 'ps') {
-	$self->{LineComment} = '%';
-	$self->{hook}=[{begin => "\n% +\n", end=>"\n% -\n",   # Reserved for output
+      $self->{LineComment} = '%';
+      #!!needs update OutDelimiters
+      $self->{hook}=[{begin => "\n% +\n", end=>"\n% -\n",   # Reserved for output
 		f=>sub{return''}},
 	       {begin => '<?', end => '!>', f => \&evaluate }];
 	$self->{CodePreparation} = 's/\\n%/\\n/g';
@@ -1061,7 +1121,7 @@ sub getmakefilelist ($$) {
 #    addHook($t1, $t2, 's/\n(?:%!)?/\n%!/g');
 #}
 
-sub echo($@) { $::O .= join('', @_) }
+sub echo :prototype(@) { $::O .= join('', @_) }
 
 # used in LaTeX mode to include verbatim textual files
 sub get_verbatim_file {
@@ -1224,8 +1284,8 @@ __END__
 
 =head1 NAME
 
-Text::Starfish.pm and starfish - A Perl-based System for Text-Embedded
-      Programming and Preprocessing
+Text::Starfish.pm and starfish - A Perl-based System for Preprocessing and
+      Text-Embedded Programming
 
 
 =head1 SYNOPSIS
@@ -1267,9 +1327,8 @@ module for embedded Perl C<Text::Oyster>, so it was not used.
 
 The idea with the `C<starfish>' name is: the Perl code is embedded into
 a text, so the text is equivalent to a shellfish containing pearls.
-A starfish comes by and eats the shellfish...  Unlike a natural
-starfish, this C<starfish> is interested in pearls and does not
-normally touch most of the surrounding meat.
+Starfish is chosen as natural starfish feeds on shellfish, with a difference
+that the system C<starfish> feeds on embedded pearls.
 
 =head1 EXAMPLES
 
@@ -1281,10 +1340,8 @@ A simple example, after running C<starfish> on a file containing:
 
 we get the following output:
 
-     <? $O= "Hello world!" !>
-     #+
-     Hello world!
-     #-
+     <? $O= "Hello world!" !>#+
+     Hello world!#-
 
 The output will not change after running the script several times.
 The same effect is achieved with:
@@ -1370,8 +1427,7 @@ Output:
   LIST=first second third\
    fourth fifth
 
-  <? echo join "\n", getmakefilelist $Star->{INFILE}, 'LIST' !>
-  #+
+  <? echo join "\n", getmakefilelist $Star->{INFILE}, 'LIST', "\n" !>#+
   first
   second
   third
@@ -1776,12 +1832,27 @@ later.  A typical usage could be as follows:
 =head2 $o->setStyle($s)
 
 Sets a particular style of the source file.  Currently implemented
-options are: html, java, latex, makefile, perl, ps, python, TeX, and
-tex.  If the parameter $s is not given, the stile given in
-$o->{STYLE} will be used if defined, otherwise it will be guessed from
-the file name in $o->{INFILE}.  If it cannot be correctly guessed, it
+options are: html, java, makefile, perl, ps, python, and tex (same as latex,
+TeX).  If the parameter $s is not given, the stile given in 
+C<$o->{STYLE}> will be used if defined, otherwise it will be guessed from
+the file name in C<$o->{INFILE}>.  If it cannot be correctly guessed, it
 will be the Perl style.
 
+Setting a style can have various side effects, but it typically
+involves setting the following variables:
+
+ $o->{Style}            # style string id
+ $o->{CodePreparation}  # function to clean the code before running
+ $o->{LineComment}      # string starting a line comment
+ $o->{OutDelimiters}    # array ref with four elements: $b1, $b2 for
+                        # starting output delimiter, and $e1, $e2 for
+                        # the ending output delimiter; $b1 and $e1
+                        # must not end with a digit, and $b2 and $e2
+                        # must not start with a digit
+ $o->{IgnoreOuter}      # boolean variable to ignore outer text, false
+                        # by default
+ $o->{hook}             # array ref, list of hooks
+
 =head1 PREDEFINED FUNCTIONS
 
 =head2 include( I<filename and options> ) -- starfish a file and echo
@@ -2020,8 +2091,7 @@ For example, in the following expansion:
  starfish: tmp
          starfish Makefile
          #<? if (-e "file.tex.sfish")
-         #{ echo "\tstarfish -o=tmp/file.tex -replace file.tex.sfish" } !>
-         #+
+         #{ echo "\tstarfish -o=tmp/file.tex -replace file.tex.sfish\n" } !>#+
          starfish -o=tmp/file.tex -replace file.tex.sfish
          #-
 
